;; Generated package description from xref.el  -*- no-byte-compile: t -*-
(define-package "xref" "1.0.4.0.20210424.95444" "Cross-referencing commands" '((emacs "26.1")) :url "https://elpa.gnu.org/packages/xref.html")
